;wfx.pbi

;justin 10/12
;Version : 0.1.2
;Functions with the hwid parameter accept both api window handle and gadget id.
;Each control created is assigned automatically a WFX_BASE object and a generic
;subclass in WM_PARENTNOTIFY.

xincludefile "wfxwindows.pbi"

enableexplicit

;Subclass API functions not included in PB (XP or higher)
prototype p_SetWindowSubclass(hWnd.i, pfnSubclass.i, uIdSubclass.i, dwRefData.i)
prototype p_RemoveWindowSubclass(hWnd.i, pfnSubclass.i, uIdSubclass.i)
prototype p_DefSubclassProc(hWnd.i, uMsg.i, wparam.i, lparam.i)

global.p_SetWindowSubclass SetWindowSubclass_
global.p_RemoveWindowSubclass RemoveWindowSubclass_
global.p_DefSubclassProc DefSubclassProc_

;- PROTOTYPES
prototype p_EventCallback(hwid.i, code.i, p1.i, p2.i, userParam.i) ;event callback for an API/PB top level or child window
prototype p_UserTopLevelProc(hwnd.i, msg.i, wparam.i, lparam.i, param.i) ;window procedure for an API top level window
prototype p_MsgLoop(hwnd.i, *msg.MSG, param.i) ;placeholder for custom message loop for an API top level window

#WFX_WC_TOPLEVEL = "*wfx_tl_wc*" ;top level window class

;Anchor flags
#WFX_ANCHORLEFT = 1
#WFX_ANCHORTOP = 2
#WFX_ANCHORRIGHT = 4
#WFX_ANCHORBOTTOM = 8
#WFX_ANCHORALL = #WFX_ANCHORLEFT | #WFX_ANCHORTOP | #WFX_ANCHORRIGHT | #WFX_ANCHORBOTTOM

;General Flags
#WFX_ANCHORREDRAW = 1 ;redraws an anchored control after moving it. default is set
#WFX_TRACKMOUSEEVENTS = 2 ;enables mouse enter, leave events. default is not set
#WFX_ANCHORCHILDS = 4 ;enables child control anchoring for the top level window. default is set

;General Notification codes
#WFX_NM_MOUSEENTER = 1 ;mouse enters top level/child window
#WFX_NM_MOUSELEAVE = 2 ;mouse exits top level/child window

;Command notification codes
#WFX_NM_CMDCLICK = 3		;control click hwid contains control handle, p1 control identifier
#WFX_NM_CMDMENU = 4			;menu selection,	p1 = menu identifier
#WFX_NM_CMDACCEL = 5		;accelerator, p1 = accelerator identifier

;Object types
#WFX_TOPLEVEL = 1
#WFX_TOPLEVEL_PB = 2
#WFX_MODAL_DIALOG = 3
#WFX_MODELESS_DIALOG = 4

;- WFX_BASE
structure WFX_BASE
	;Public
	hwnd.i
	pbid.i ;only used to store the window id for top level pb windows
	type.w 
	param.i ;user param
	flags.l ;General flags
	anchor.l ;Anchor flags
	*callBack.p_EventCallback
	callBackParam.i

	;Private
	idSubclass.i
	wfx__OldProc.i ;Used for subclassing
	mouseOver.b ;used to track mouse events
	leftmargin.l ;anchoring
	topmargin.l
	rightmargin.l
	bottommargin.l
endstructure 

;- WFX_TOPLEVEL
structure WFX_TOPLEVEL extends WFX_BASE
 	*userProc.p_UserTopLevelProc
	;userProcParam.i
	hAccelTable.i ;win Accelerator table handle
endstructure

structure WFX_CREATEDATA
	*userProc.p_UserTopLevelProc
	userProcParam.i
endstructure 

;- DECLARES
declare wfx__ChildProc(hwnd.i, msg.i, wparam.i, lparam.i)
declare wfx__EnumTopLevelChilds(hwchild.i, param.i)
declare wfx__FreeBaseObject(hwnd.i)
declare wfx__FreeTopLevelObject(hwnd.i)
declare wfx__Init()
declare wfx__TopLevelProc(hwnd.i, msg.i, wparam.i, lparam.i)
declare wfx__SubClass(hwnd.i, *proc.i, param.i)
declare wfx__OnParentNotify(hwnd.i, msg.i, wparam.i, lparam.i, notDialog.b=#true)
declare wfx__MakeTopLevelObject(type.w)
declare wfx__MakeBaseObject(size.i=sizeof(WFX_BASE))
declare wfx__GetChildPos(hwnd.i, *rc.RECT)

;- INIT
wfx__Init()

;- BASE FUNCTIONS
;Sets and gets the internal window object
macro wfx__SetWindowObject(hwnd, obj)
	SetWindowLongPtr_(hwnd,  #GWLP_USERDATA, obj)
endmacro

macro wfx__GetWindowObject(hwnd)
	GetWindowLongPtr_(hwnd,  #GWLP_USERDATA)
endmacro

macro wfx__GetTopLevelWindowObject(hwchild)
	wfx__GetWindowObject(GetAncestor_(hwchild, #GA_ROOT))
endmacro

macro wfx__SetDefaultFont(hwnd)
;Sets the default gui font for a control
	SendMessage_(hwnd, #WM_SETFONT, GetStockObject_(#DEFAULT_GUI_FONT), #true)
endmacro

;Puts, removes or checks if a window has a style or extended style
macro wfx__PutStyle(hwnd, style, type=#GWL_STYLE)
	SetWindowLongPtr_(hwnd, type, GetWindowLongPtr_(hwnd, type) | style)
endmacro

macro wfx__RemoveStyle(hwnd, style, type=#GWL_STYLE)
	SetWindowLongPtr_(hwnd, type, GetWindowLongPtr_(hwnd, type) & ~style)
endmacro 

procedure wfx__HasStyle(hwnd.i, style.l, type.i=#GWL_STYLE)
	if GetWindowLongPtr_(hwnd, type) & style = style
		procedurereturn #true
	else
		procedurereturn #false
	endif
endprocedure 

;Puts, removes or checks if a window has a wfx flag
procedure wfx__HasFlag(hwnd.i, flag.l)
	define.WFX_BASE *this = wfx__GetWindowObject(hwnd)

	if *this
		if *this\flags & flag = flag
			procedurereturn #true
		else
			procedurereturn #false
		endif
	endif 
endprocedure

procedure wfx__PutFlag(hwnd.i, flag.l)
	define.WFX_BASE *this = wfx__GetWindowObject(hwnd)

	if *this : *this\flags = *this\flags | flag : endif
endprocedure

procedure wfx__RemoveFlag(hwnd.i, flag.l)
	define.WFX_BASE *this = wfx__GetWindowObject(hwnd)

	if *this : *this\flags = *this\flags & ~flag : endif
endprocedure

procedure wfx__MakeBaseObject(size.i=sizeof(WFX_BASE))
	define.WFX_BASE *this

	*this = allocatememory(size)
	*this\flags = #WFX_ANCHORREDRAW

	procedurereturn *this
endprocedure

procedure wfx__MakeTopLevelObject(type.w)
	define.WFX_TOPLEVEL *this

	*this = allocatememory(sizeof(WFX_TOPLEVEL))
	*this\flags = #WFX_ANCHORCHILDS
	*this\type = type

	procedurereturn *this
endprocedure  

procedure wfx__FreeTopLevelObject(hwnd.i)
	define.WFX_TOPLEVEL *this = wfx__GetWindowObject(hwnd)
	
	if *this
		freememory(*this)
	endif 
endprocedure 

procedure wfx__FreeBaseObject(hwnd.i)
	define.WFX_BASE *this = wfx__GetWindowObject(hwnd)

	if *this
		freememory(*this)
	endif
endprocedure

procedure wfx__RegisterClass()
;Registers the top level wfx window class

	define.WNDCLASSEX wc
	define.s class

	class = #WFX_WC_TOPLEVEL
	wc\cbSize = sizeof(WNDCLASSEX)
	wc\lpfnWndProc = @wfx__TopLevelProc()
	wc\hInstance = GetModuleHandle_(0)
	wc\hCursor = LoadCursor_(#null, #IDC_ARROW)
	wc\hbrBackground = #COLOR_BTNFACE + 1
	wc\lpszClassName = @class

	procedurereturn RegisterClassEx_(@wc)
endprocedure 

procedure wfx__LoadComctlFunctions()
;Loads Comctl functions not natively supported by PB

	define.i hComCtl32

	hComCtl32 = LoadLibrary_("comctl32.dll")
	if hComCtl32
		;Subclass functions for XP or higher
		SetWindowSubclass_ = GetProcAddress_(hComCtl32, MAKELONG(410,0))
		RemoveWindowSubclass_ = GetProcAddress_(hComCtl32, MAKELONG(412,0))
		DefSubclassProc_ = GetProcAddress_(hComCtl32, MAKELONG(413,0))
		FreeLibrary_(hComCtl32)

		if SetWindowSubclass_ and RemoveWindowSubclass_ and DefSubclassProc_
			procedurereturn #true
		else
			procedurereturn #false
		endif 

	else
		procedurereturn #false
	endif 
endprocedure 

procedure wfx__Init()
	if OSVersion()>=#PB_OS_Windows_XP
		wfx__LoadComctlFunctions()
	endif 

	wfx__RegisterClass()
endprocedure

procedure wfx__DefSubclassProc(hwnd.i, msg.i, wparam.i, lparam.i, *this.WFX_TOPLEVEL, notDialog.b=#true)	
;Calls the PB subclass procedure, the user API procedure or the default procedure

	if *this 
 		if *this\wfx__OldProc ;PB window
 			procedurereturn CallWindowProc_(*this\wfx__OldProc, hwnd, msg, wparam, lparam)

		elseif *this\userProc ;API window / dialog
			procedurereturn *this\userProc(hwnd, msg, wparam, lparam, *this\callBackParam)
		endif 
	
	else ;default processing
		if notDialog
			procedurereturn DefWindowProc_(hwnd, msg, wparam, lparam)
		else ;dialog
			procedurereturn #false
		endif 
	endif 
endprocedure 

procedure wfx__OpenWindow(exstyle.l, title.s, style.l, x.i, y.i, width.i, height.i, hmenu.i, *proc.p_UserTopLevelProc, param.i=#null, *msgLoop.p_MsgLoop=#null, *wc.WNDCLASSEX=#null)
;Opens an API window and starts a basic message loop, you can use your own message loop using the
;*msgLoop callback. 'param' is passed as the last additional parameter to your window procedure
;and to all the child control event callback.
;The function returns when you exit the message loop calling PostQuitMessage_() in your window procedure.
;You can pass your own window class structure that will be registered, the lpfnWndProc field will be replaced
;by the wfx procedure.

	define.i hwnd
	define.MSG winMsg
	define.WFX_CREATEDATA cd
	define.s class
	define.WFX_TOPLEVEL *this

	if *proc=#null : procedurereturn : endif 

	class = #WFX_WC_TOPLEVEL
	if *wc
		*wc\lpfnWndProc = @wfx__TopLevelProc()
		if *wc\lpszClassName
			class = peeks(*wc\lpszClassName)
		endif
		RegisterClassEx_(*wc)
	endif 

	cd\userProc = *proc
	cd\userProcParam = param
	hwnd = CreateWindowEx_(exstyle, class, title, style, x, y, width, height, 0, hmenu, GetModuleHandle_(0), @cd)
	if hwnd
		*this = wfx__GetWindowObject(hwnd)
		UpdateWindow_(hwnd)

		if *msgLoop ;Custom message loop
			*msgLoop(hwnd, @winMsg, param)

		else
			while GetMessage_(@winMsg, #Null, 0, 0 )<>0
				if TranslateAccelerator_(hwnd, *this\hAccelTable, @winMsg)=0 and IsDialogMessage_(hwnd, @winMsg)=0  
					TranslateMessage_(@winMsg)
					DispatchMessage_(@winMsg)  
				endif 
			wend
		endif 
	endif 

	procedurereturn winMsg\wParam ;exit code
endprocedure

procedure wfx__GetHwnd(hwid.i)
;Gets the window api handle from a pb id or an api handle

	if Iswindow(hwid)
		procedurereturn windowid(hwid)
	elseif isgadget(hwid)
		procedurereturn gadgetid(hwid)
	elseif IsWindow_(hwid)
		procedurereturn hwid
	endif
endprocedure

macro wfx__IsGadget(hwnd)
	IsGadget(GetProp_(hwnd, "PB_ID"))
endmacro 

procedure wfx__HwndToPbGadgetId(hwnd.i)
;Returns the PB Gadget ID from a window handle or the window handle if it is not a PB Gadget
	define.i pbid

	pbid = GetProp_(hwnd, "PB_ID")
	if IsGadget(pbid)
		procedurereturn pbid
	else
		procedurereturn hwnd
	endif 
endprocedure 

procedure.s wfx__GetClassName(hwnd.i)
	define.i buf
	define.s text

	buf = allocatememory(#MAX_PATH)
	GetClassName_(hwnd, buf, #MAX_PATH)
	text = peeks(buf)
	freememory(buf)

	procedurereturn text
endprocedure 

procedure wfx__RemoveWindowTheme(hwnd.i)
	define.s nstr=""
	
  SetWindowTheme_(hwnd, #Null, @nstr)
endprocedure 

procedure wfx__GetChildPos(hwnd.i, *rc.RECT)
;Gets the child coordinates relative to its parent

	GetWindowRect_(hwnd, *rc)
	ScreenToClient_(GetParent_(hwnd), *rc)
	ScreenToClient_(GetParent_(hwnd), @*rc\right)
endprocedure 

procedure wfx__SubClass(hwnd.i, *proc.i, param.i)
;Installs a window subclass

	define.WFX_BASE *this

	*this = wfx__GetWindowObject(hwnd)
	if *this
		*this\idSubclass = *this\idSubclass + 1
		if SetWindowSubclass_(hwnd, *proc, *this\idSubclass, param)
			procedurereturn *this\idSubclass
		else
			procedurereturn -1
		endif 
	endif 
endprocedure 


procedure wfx__SetAnchor(hwnd.i, anchorflags.l)
;Sets achor flags

	define.WFX_BASE *this = wfx__GetWindowObject(hwnd)

	if *this : *this\anchor = anchorflags : endif 
endprocedure 

procedure wfx__GetAnchor(hwnd.i)
;Gets anchor flags
	define.WFX_BASE *this = wfx__GetWindowObject(hwnd)

	if *this
		procedurereturn *this\anchor
	endif 
endprocedure 

procedure wfx__GetParam(hwnd.i)
;Gets user param for a top level window, dialog or control

	define.WFX_BASE *this = wfx__GetWindowObject(hwnd)

	if *this
		procedurereturn *this\param
	endif
endprocedure

procedure wfx__SetParam(hwnd.i, param.i)
;Sets user param for a top level window, dialog or control

	define.WFX_BASE *this = wfx__GetWindowObject(hwnd)

	if *this
		*this\param = param
	endif
endprocedure

procedure wfx__SetUpChildControl(hwchild.i)
;Sets a base window object, installs the base subclass and saves initial position

	define.RECT rcParent, rcChild
	define.WFX_BASE *child
	define.WFX_TOPLEVEL *top = wfx__GetTopLevelWindowObject(hwchild)

	*child = wfx__MakeBaseObject()
	if *top ;save callback param
		*child\callBackParam = *top\callBackParam
	endif 
	*child\hwnd = hwchild
	wfx__SetWindowObject(hwchild, *child)
	*child\wfx__OldProc = SetWindowLongPtr_(hwchild, #GWLP_WNDPROC, @wfx__ChildProc()) ;subclass

	;Save child position
	GetClientRect_(GetParent_(hwchild), @rcParent)
	wfx__GetChildPos(hwchild, @rcChild)
	with *child
		\leftmargin = rcChild\left
		\topmargin = rcChild\top
		\rightmargin = rcParent\right - \leftmargin - (rcChild\right - rcChild\left) ;childWidth
		\bottommargin = rcParent\bottom - \topmargin - (rcChild\bottom - rcChild\top) ;childHeight
	endwith

	procedurereturn *child
endprocedure  

procedure wfx__SetCallback(hwnd.i, *callback.p_EventCallback)
;Sets the event callback for this top level or child window

	define.WFX_BASE *this = wfx__GetWindowObject(hwnd)

	*this\callBack = *callback
endprocedure

procedure wfx__GetCallBack(hwnd.i)
;Gets the event callback

	define.WFX_BASE *this = wfx__GetWindowObject(hwnd)

	procedurereturn *this\callBack
endprocedure 

procedure wfx__TrackMouse(hwnd.i, flags.l, hoverTime.l)
;Wraps the TrackMouseEvent_() function

	define.TRACKMOUSEEVENT tme
	
	tme\cbSize = SizeOf(TRACKMOUSEEVENT)
	tme\dwFlags = flags
	tme\hwndTrack = hwnd
	tme\dwHoverTime = hoverTime
	
	procedurereturn TrackMouseEvent_(@tme)
endprocedure 

procedure wfx__ExtendWindowObject(hwnd.i, newsize.i)
;Extends an existing window object to a new size preserving the old contents, used to
;extend the base object when creating new controls

	define.i *newObject, *oldObject

	*oldObject = wfx__GetWindowObject(hwnd)
	if *oldObject
		*newObject = reallocatememory(*oldObject, newsize)
		wfx__SetWindowObject(hwnd, *newObject)

		procedurereturn *newObject
	endif
endprocedure 

procedure wfx__CallEvent(*this.WFX_BASE, code.i, p1.i, p2.i)
;Calls an event for a top level/child window, for API windows passes the window handle,
;for PB windows the PB ID

	if *this and *this\callBack
		if *this\type=#WFX_TOPLEVEL or *this\type=#WFX_MODAL_DIALOG or *this\type=#WFX_MODELESS_DIALOG ;api window / dialog
			*this\callBack(*this\hwnd, code, p1, p2, *this\callBackParam)
	
		elseif *this\type=#WFX_TOPLEVEL_PB ;pb window
			procedurereturn *this\callBack(*this\pbid, code, p1, p2, *this\callBackParam)
	
		elseif wfx__IsGadget(*this\hwnd)
			procedurereturn *this\callBack(wfx__HwndToPbGadgetId(*this\hwnd), code, p1, p2, *this\callBackParam)

		else ;api control
			procedurereturn *this\callBack(*this\hwnd, code, p1, p2, *this\callBackParam)
		endif
	endif 
endprocedure 

procedure wfx__SetAccelTable(hwnd.i, haccel.i)
;Sets the accelerator table for a top level window

	define.WFX_TOPLEVEL *this = wfx__GetWindowObject(hwnd)

	if *this : *this\hAccelTable = haccel : endif
endprocedure

procedure wfx__GetAccelTable(hwnd.i)
;Gets the current accelerator table of the top level window

	define.WFX_TOPLEVEL *this = wfx__GetWindowObject(hwnd)

	if *this : procedurereturn *this\hAccelTable : endif
endprocedure

procedure wfx__EnumTopLevelChilds(hwchild.i, hwParent.i)
;Positions anchored controls, code by srod

	define.WFX_BASE *child 
	define.RECT rcParent, rcChild
	define.i l, t, r, b, childWidth, childHeight
	define.b redraw

	*child = wfx__GetWindowObject(hwchild)
	if *child and *child\anchor
		wfx__GetChildPos(hwChild, @rcChild)

		l = rcChild\left : t = rcChild\top : r = rcChild\right : b = rcChild\bottom
		childWidth = rcChild\right - rcChild\left
		childHeight = rcChild\bottom - rcChild\top

		GetClientRect_(GetParent_(hwChild), @rcParent)
		if *child\anchor & #WFX_ANCHORRIGHT
			r = rcParent\right - *child\rightmargin
		endif

		if *child\anchor & #WFX_ANCHORLEFT = 0
			l = r - childWidth
		endif

		if *child\anchor & #WFX_ANCHORBOTTOM
			b = rcParent\bottom - *child\bottommargin
		endif

		if *child\anchor & #WFX_ANCHORTOP = 0
			t = b - childHeight
		endif

		if *child\flags & #WFX_ANCHORREDRAW = #WFX_ANCHORREDRAW
			redraw = #true
		else
			redraw = #false
		endif 

		MoveWindow_(hwchild, l, t, r-l, b-t, redraw)
	endif

	procedurereturn #true
endprocedure 

;- GENERIC MESSAGE HANDLERS
;Used by both the top level window and child controls,
;handlers using the notDialog param are also common to dialogs.

procedure wfx__OnParentNotify(hwnd.i, msg.i, wparam.i, lparam.i, notDialog.b=#true)
;Sets child window object and subclass

 	define.WFX_BASE *this = wfx__GetWindowObject(hwnd)
	define.WFX_BASE *child

	if LOWORD(wparam)=#WM_CREATE
 		if wfx__GetWindowObject(lparam)=#null
 			*child = wfx__SetUpChildControl(lparam)
		endif 
 	endif 

	procedurereturn wfx__DefSubclassProc(hwnd, msg, wparam, lparam, *this, notDialog)
endprocedure 

procedure wfx__OnMouseMove(hwnd.i, msg.i, wParam.i, lParam.i, notDialog.b=#true)
;Used to catch mouseenter, mouseleave events

 	define.WFX_BASE *this = wfx__GetWindowObject(hwnd)

	if *this\flags & #WFX_TRACKMOUSEEVENTS = #WFX_TRACKMOUSEEVENTS
		if (not *this\mouseOver)  
			wfx__TrackMouse(hwnd, #TME_HOVER|#TME_LEAVE, 1)
		endif
	endif  
  
	procedurereturn wfx__DefSubclassProc(hwnd, msg, wparam, lparam, *this, notDialog)
endprocedure 

procedure wfx__OnMouseHover(hwnd.i, msg.i, wParam.i, lParam.i, notDialog.b=#true)
;Used to catch mouseenter, mouseleave events

 	define.WFX_BASE *this = wfx__GetWindowObject(hwnd)

	*this\mouseOver = #True 

	if *this\callBack and *this\flags & #WFX_TRACKMOUSEEVENTS = #WFX_TRACKMOUSEEVENTS
		wfx__CallEvent(*this, #WFX_NM_MOUSEENTER, 0, 0)
	endif 

	procedurereturn wfx__DefSubclassProc(hwnd, msg, wparam, lparam, *this, notDialog)
endprocedure 

procedure wfx__OnMouseLeave(hwnd.i, msg.i, wParam.i, lParam.i, notDialog.b=#true)
;Used to catch mouseenter, mouseleave events

 	define.WFX_BASE *this = wfx__GetWindowObject(hwnd)

	*this\mouseOver = #False 
	wfx__TrackMouse(hwnd, #TME_HOVER, 1)

	if *this\callBack and *this\flags & #WFX_TRACKMOUSEEVENTS = #WFX_TRACKMOUSEEVENTS
		wfx__CallEvent(*this, #WFX_NM_MOUSELEAVE, 0, 0)
	endif 

	procedurereturn wfx__DefSubclassProc(hwnd, msg, wparam, lparam, *this, notDialog)
endprocedure 

procedure wfx__OnCommand(hwnd.i, msg.i, wParam.i, lParam.i, notDialog.b=#true)
 	define.WFX_BASE *this = wfx__GetWindowObject(hwnd)

	if lparam ;is control
		wfx__CallEvent(wfx__GetWindowObject(lParam), #WFX_NM_CMDCLICK, LOWORD(wparam), 0)
	else 
		if HIWORD(wparam) ;accelerator
			wfx__CallEvent(*this, #WFX_NM_CMDACCEL, LOWORD(wparam), 0)
		else ;menu
			wfx__CallEvent(*this, #WFX_NM_CMDMENU, LOWORD(wparam), 0)
		endif 
	endif 

	procedurereturn wfx__DefSubclassProc(hwnd, msg, wparam, lparam, *this, notDialog)
endprocedure 

;- TOP LEVEL MESSAGE HANDLERS
procedure wfx__OnTopLevelCreate(hwnd.i, msg.i, wparam.i, *cs.CREATESTRUCT)
;Sets window object, only handled by API windows

	define.WFX_CREATEDATA *cd
	define.WFX_TOPLEVEL *this

	*this = wfx__MakeTopLevelObject(#WFX_TOPLEVEL)
	*cd = *cs\lpCreateParams

	*this\userProc = *cd\userProc
	*this\callBackParam = *cd\userProcParam
	*this\hwnd = hwnd
	wfx__SetWindowObject(hwnd, *this)

	procedurereturn wfx__DefSubclassProc(hwnd, msg, wparam, *cs, *this)
endprocedure 

procedure wfx__OnTopLevelSize(hwnd.i, msg.i, wparam.i, lparam.i, notDialog.b=#true)
;Positions anchored controls

	define.WFX_TOPLEVEL *this = wfx__GetWindowObject(hwnd)

	if *this\flags & #WFX_ANCHORCHILDS = #WFX_ANCHORCHILDS
  	EnumChildWindows_(hwnd,  @wfx__EnumTopLevelChilds(), hwnd)
	endif

	procedurereturn wfx__DefSubclassProc(hwnd, msg, wparam, lparam, *this, notDialog)
endprocedure 

procedure wfx__OnTopLevelDestroy(hwnd.i, msg.i, wparam.i, lparam.i)
;Removes sublass and frees window object

	define.WFX_TOPLEVEL *this = wfx__GetWindowObject(hwnd)
	define.i ret

	if *this\wfx__OldProc ;PB window, remove subclass
		SetWindowLongPtr_(hwnd, #GWLP_WNDPROC, *this\wfx__OldProc)
	endif 

	ret = wfx__DefSubclassProc(hwnd, msg, wparam, lparam, *this)	

	wfx__FreeTopLevelObject(hwnd)
	*this = #null

	procedurereturn ret
endprocedure 

procedure wfx__TopLevelProc(hwnd.i, msg.i, wparam.i, lparam.i)
	select msg
		case #WM_CREATE : procedurereturn wfx__OnTopLevelCreate(hwnd, msg, wparam, lparam)

		case #WM_COMMAND : procedurereturn wfx__OnCommand(hwnd, msg, wParam, lParam)

 		case #WM_PARENTNOTIFY : procedurereturn wfx__OnParentNotify(hwnd, msg, wparam, lparam)

    case #WM_SIZE : procedurereturn wfx__OnTopLevelSize(hwnd, msg, wparam, lparam)

		case #WM_MOUSEMOVE : procedurereturn wfx__OnMouseMove(hwnd, msg, wparam, lparam)

		case #WM_MOUSEHOVER : procedurereturn wfx__OnMouseHover(hwnd, msg, wparam, lparam)
	
		case #WM_MOUSELEAVE : procedurereturn wfx__OnMouseLeave(hwnd, msg, wparam, lparam)

		case #WM_DESTROY : procedurereturn wfx__OnTopLevelDestroy(hwnd, msg, wparam, lparam)

		default : procedurereturn wfx__DefSubclassProc(hwnd, msg, wparam, lparam, wfx__GetWindowObject(hwnd))
	endselect 
endprocedure 

;- CHILD MESSAGE HANDLERS
procedure wfx__OnChildWindowPosChanged(hwnd.i, msg.i, wparam.i, *wp.WINDOWPOS)
;Updates child margins

	define.WFX_BASE *this = wfx__GetWindowObject(hwnd)
	define.RECT rcParent

	if *this\anchor and *wp\cx and *wp\cy
		GetClientRect_(GetParent_(hwnd), @rcParent)
		With *this
			\leftmargin = *wp\x
			\topmargin = *wp\y
			\rightmargin = rcParent\right - \leftmargin - *wp\cx
			\bottommargin = rcParent\bottom - \topmargin - *wp\cy
		EndWith
	endif 

	procedurereturn wfx__DefSubclassProc(hwnd, msg, wparam, *wp, *this)
endprocedure

procedure wfx__OnChildDestroy(hwnd.i, msg.i, wparam.i, lparam.i)
;Removes subclass and frees window object

	define.WFX_BASE *this = wfx__GetWindowObject(hwnd)

	if *this\wfx__OldProc
		SetWindowLongPtr_(hwnd, #GWLP_WNDPROC, *this\wfx__OldProc)
	endif

	wfx__FreeBaseObject(hwnd)

	procedurereturn DefWindowProc_(hwnd, msg, wparam, lparam)
endprocedure 

procedure wfx__ChildProc(hwnd.i, msg.i, wparam.i, lparam.i)
 	select msg
		case #WM_PARENTNOTIFY : procedurereturn wfx__OnParentNotify(hwnd, msg, wparam, lparam)

		case #WM_COMMAND : procedurereturn wfx__OnCommand(hwnd, msg, wParam, lParam)

		case #WM_WINDOWPOSCHANGED : procedurereturn wfx__OnChildWindowPosChanged(hwnd, msg, wparam, lparam)

		case #WM_MOUSEMOVE : procedurereturn wfx__OnMouseMove(hwnd, msg, wparam, lparam)

		case #WM_MOUSEHOVER : procedurereturn wfx__OnMouseHover(hwnd, msg, wparam, lparam)
	
		case #WM_MOUSELEAVE : procedurereturn wfx__OnMouseLeave(hwnd, msg, wparam, lparam)

		case #WM_DESTROY : procedurereturn wfx__OnChildDestroy(hwnd, msg, wparam, lparam)

		default : procedurereturn wfx__DefSubclassProc(hwnd, msg, wparam, lparam, wfx__GetWindowObject(hwnd))
	endselect 
endprocedure

procedure wfx_SetUpPBWindow(id.i, param.i=0)
;Sets top level window object and subclass, must be called on main PB window before
;adding gadgets to enable wfx freatures. Accepts only a PB Window ID.

	define.WFX_TOPLEVEL *this
	define.i hwnd

	if iswindow(id)=0 : procedurereturn : endif 

	hwnd = windowid(id)

	*this = wfx__MakeTopLevelObject(#WFX_TOPLEVEL_PB)
	*this\hwnd = hwnd
	*this\pbid = id
	*this\callBackParam = param
	wfx__SetWindowObject(hwnd, *this)

	;Subclass
	*this\wfx__OldProc = SetWindowLongPtr_(hwnd, #GWLP_WNDPROC, @wfx__TopLevelProc())
endprocedure 

;- PB MACROS
macro wfx_SetAnchor(hwid, anchorflags)
	wfx__SetAnchor(wfx__GetHwnd(hwid), anchorflags)
endmacro 

macro wfx_GetAnchor(hwid)
	wfx__GetAnchor(wfx__GetHwnd(hwid))
endmacro

macro wfx_GetParam(hwid)
	wfx__GetParam(wfx__GetHwnd(hwid))
endmacro

macro wfx_SetParam(hwid, param)
	wfx__SetParam(wfx__GetHwnd(hwid), param)
endmacro 

macro wfx_HasFlag(id, flag) 
	wfx__HasFlag(wfx__GetHwnd(hwid), flag)
endmacro

macro wfx_PutFlag(hwid, flag)
	wfx__PutFlag(wfx__GetHwnd(hwid), flag)
endmacro

macro wfx_RemoveFlag(hwid, flag)
	wfx__RemoveFlag(wfx__GetHwnd(hwid), flag)
endmacro

macro wfx_PutStyle(hwid, style, type=#GWL_STYLE)
	wfx__PutStyle(wfx__GetHwnd(hwid), style, type)
endmacro

macro wfx_HasStyle(hwid, style, type=#GWL_STYLE)
	wfx__HasStyle(wfx__GetHwnd(hwid), style, type)
endmacro

macro wfx_RemoveStyle(hwid, style, type=#GWL_STYLE)
	wfx__RemoveStyle(wfx__GetHwnd(hwid), style, type)
endmacro

macro wfx_RemoveWindowTheme(hwid)
	wfx__RemoveWindowTheme(wfx__GetHwnd(hwid))
endmacro

macro wfx_Subclass(hwid, proc, param)
	wfx__SubClass(wfx__GetHwnd(hwid), proc, param)
endmacro

macro wfx_SetCallback(hwid, callback)
	wfx__SetCallback(wfx__GetHwnd(hwid), callback)
endmacro

macro wfx_GetCallback(hwid)
	wfx__GetCallback(wfx__GetHwnd(hwid))
endmacro












