;wfxDialogs.pbi

;Functions to create empty dialogs using memory templates, the dialog controls should be created
;in the #WM_INITDIALOG message.

;From MSDN:
;Modeless Dialogs:
;The dialog box template should specify the #WS_POPUP, #WS_CAPTION, #WS_BORDER,
;and WS_SYSMENU styles. The system does not automatically display the dialog box unless the template
;specifies the #WS_VISIBLE style.
;Use DestroyWindow_() to close the dialog.
;
;Modal Dialogs:
;The dialog box template should specify the #WS_POPUP, #WS_SYSMENU, #WS_CAPTION, and #DS_MODALFRAME styles.
;The system always displays a modal dialog box regardless of whether the dialog box template specifies
;the #WS_VISIBLE style.
;Use EndDialog_() to close the dialog.
;
;Dialog Box Procedure:
;Returns #TRUE if it processes a message or #FALSE if it does not.
;
;Keyboard interface:
;The system automatically provides the keyboard interface for all modal dialog boxes.
;It does not provide the interface for modeless dialog boxes unless the application calls the IsDialogMessage
;function to filter messages in its main message loop.

xincludefile "wfx.pbi"

;- PROTOTYPES
prototype p_DlgUserProc(hwnd.i, msg.i, wparam.i, lparam.i, userparam.i)

structure WFX_DLGTEMPLATE extends DLGTEMPLATE
	menu.w
	class.w
	title.w
endstructure 

structure WFX_DLG_INITDATA
	*tmp._DLGTEMPLATE
	userProcParam.i
	*userProc.i
	type.w ;MODAL / MODELESS
endstructure

structure WFX_DIALOG extends WFX_TOPLEVEL

endstructure

;- DECLARES
declare wfx__DlgProc(hwnd.i, msg.i, wparam.i, lparam.i)

;- BASE FUNCTIONS
procedure wfx__MakeDialogObject(type.w) 
	define.WFX_DIALOG *this
	
	*this = allocatememory(sizeof(WFX_DIALOG))
	*this\type = type
	*this\flags = #WFX_ANCHORCHILDS
		
	procedurereturn *this
endprocedure 

procedure wfx__FreeDialogObject(hwnd.i)
	define.WFX_DIALOG *this = wfx__GetWindowObject(hwnd)

	if *this
		freememory(*this)
	endif
endprocedure 

procedure wfx__CloseDialog(hwnd.i, modalResult.i=0)
;Closes a modal / modeless dialog

	define.WFX_DIALOG *this = wfx__GetWindowObject(hwnd)

	if *this
		if *this\type=#WFX_MODAL_DIALOG
			procedurereturn EndDialog_(hwnd, modalResult)

		elseif *this\type=#WFX_MODELESS_DIALOG
			procedurereturn DestroyWindow_(hwnd)
		endif 
	endif 
endprocedure 

procedure wfx__MakeDlgTemplate(style.l, x.w, y.w, width.w, height.w, title.s)
;Creates a memory dialog template with no controls

	define.i tmpSize, titleLen
	define.WFX_DLGTEMPLATE *tmp
		
	if title=""
		titleLen = 1
	else
		titleLen = (len(title) * 2) + 2
	endif 
	
	tmpSize = sizeof(WFX_DLGTEMPLATE) + titleLen - 1 
	tmpSize + (tmpSize % sizeof(WORD))
	*tmp = allocatememory(tmpSize)
	
	*tmp\style = style
	*tmp\x = x
	*tmp\y = y
	*tmp\cx = width
	*tmp\cy = height
	*tmp\cdit = 0
	*tmp\menu = 0
	*tmp\class = 0
	*tmp\title = 0
	
	if title<>""
		pokes(@*tmp\title, title, titleLen - 2, #PB_Unicode)
	endif 
	
	procedurereturn *tmp
endprocedure 

procedure wfx__DlgOpenModal(hwparent.i, style.l, x.w, y.w, width.w, height.w, title.s, *dlgproc.p_DlgUserProc, userParam.i=0)
;Opens a modal dialog, userparam is passed as an aditional parameter to the dialog procedure

	define.WFX_DLG_INITDATA initDat

	if *dlgproc=0 : procedurereturn : endif 
	
	initDat\tmp = wfx__MakeDlgTemplate(style, x, y, width, height, title)
	initDat\userProcParam = userParam
	initDat\userProc = *dlgproc
	initDat\type = #WFX_MODAL_DIALOG
	
	procedurereturn DialogBoxIndirectParam_(GetModuleHandle_(0), initDat\tmp, hwparent, @wfx__DlgProc(), @initDat)
endprocedure

procedure wfx__DlgOpenModeless(hwparent.i, style.l, x.w, y.w, width.w, height.w, title.s, *dlgproc.p_DlgUserProc, userParam.i=0)
;Opens a modeless dialog, userparam is passed as an aditional parameter to the dialog procedure

	define.WFX_DLG_INITDATA initDat

	if *dlgproc=0 : procedurereturn : endif 
	
	initDat\tmp = wfx__MakeDlgTemplate(style, x, y, width, height, title)
	initDat\userProcParam = userParam
	initDat\userProc = *dlgproc
	initDat\type = #WFX_MODELESS_DIALOG

	procedurereturn CreateDialogIndirectParam_(GetModuleHandle_(0), initDat\tmp, hwparent, @wfx__DlgProc(), @initDat)
endprocedure

procedure wfx__DefDialogProc(hwnd, msg, wparam, lparam)
	define.WFX_DIALOG *this = wfx__GetWindowObject(hwnd)

	if *this
		procedurereturn *this\userProc(hwnd, msg, wparam, lparam, *this\callbackParam)
	else
		procedurereturn #false
	endif 
endprocedure  

;- MESSAGE HANDLERS
procedure wfx__OnDlgInitDialog(hwnd.i, msg.i, wparam.i, *dat.WFX_DLG_INITDATA)
;Sets dialog object

	define.WFX_DIALOG *this

	*this = wfx__MakeDialogObject(*dat\type)

	*this\userProc = *dat\userProc
	*this\callBackParam = *dat\userProcParam
	*this\hwnd = hwnd

	wfx__SetWindowObject(hwnd, *this)

	freememory(*dat\tmp) ;Free template memory

	procedurereturn *this\userProc(hwnd, msg, wparam, *this\callBackParam, *this\callBackParam)
endprocedure

procedure wfx__OnDlgDestroy(hwnd.i, msg.i, wparam.i, lparam.i)
;Frees dialog object

	define.WFX_DIALOG *this = wfx__GetWindowObject(hwnd)
	define.i ret

	ret = *this\userProc(hwnd, msg, wparam, lparam, *this\callBackParam)
	wfx__FreeDialogObject(hwnd)

	procedurereturn ret
endprocedure 

procedure wfx__DlgProc(hwnd.i, msg.i, wparam.i, lparam.i)
;Preprocesses some messages before calling the user procedure 

	select msg
		case #WM_INITDIALOG : procedurereturn wfx__OnDlgInitDialog(hwnd, msg, wparam, lparam)

 		case #WM_PARENTNOTIFY : procedurereturn wfx__OnParentNotify(hwnd, msg, wparam, lparam, #false)

		case #WM_COMMAND : procedurereturn wfx__OnCommand(hwnd, msg, wParam, lParam, #false)
 
 		case #WM_MOUSEMOVE : procedurereturn wfx__OnMouseMove(hwnd, msg, wparam, lparam, #false)
 
 		case #WM_MOUSEHOVER : procedurereturn wfx__OnMouseHover(hwnd, msg, wparam, lparam, #false)
 	
 		case #WM_MOUSELEAVE : procedurereturn wfx__OnMouseLeave(hwnd, msg, wparam, lparam, #false)

    case #WM_SIZE : procedurereturn wfx__OnTopLevelSize(hwnd, msg, wparam, lparam, #false)

		case #WM_DESTROY : procedurereturn wfx__OnDlgDestroy(hwnd, msg, wparam, lparam)

		default : procedurereturn wfx__DefDialogProc(hwnd, msg, wparam, lparam)
	endselect 
endprocedure

;- PB MACROS
macro wfx_DlgOpenModal(hwid, style, x, y, width, height, title, dlgproc, userParam=0)
	wfx__DlgOpenModal(wfx__GetHwnd(hwid), style, x, y, width, height, title, dlgproc, userParam)
endmacro

macro wfx_DlgOpenModeless(hwid, style, x, y, width, height, title, dlgproc, userParam=0)
	wfx__DlgOpenModeless(wfx__GetHwnd(hwid), style, x, y, width, height, title, dlgproc, userParam)
endmacro







